/* * 
 * This file is part of bufferpool
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 * 
 * bufferpool is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * bufferpool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with bufferpool; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA 
 *
 * */

/**
 * @file bp.c
 * Bufferpool implementation
 */

#include <sys/mman.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "bp_internal.h"
#include "bufferpool.h"

#ifndef PATH_MAX
/**
 * @TODO This definition should be removed and bufferpool be
 * independent from this; the new versions of glibc and Linux don't
 * impose this restriction.
 */
#define PATH_MAX 4096
#endif

#define omsbuff_min(x,y) ((x) < (y) ? (x) : (y))

/**
 * @brief internal function used to add a page of slots of BPBUFF_MEM_PAGE
 * size.
 *
 * @warning the function assumes that the caller (bp_write or
 * bp_getslot) locked the buffer mutex
 *
 * @return the first BPSlot of new added page of slots.
 */

BPSlot *bp_addpage(BPBuffer * buffer, BPSlot * prev)
{
    BPSlot *added;
    unsigned int i;
    ptrdiff_t prev_diff;

    switch (buffer->type) {
    case buff_shm:
        prev_diff = prev - buffer->slots;
        added = bp_shm_addpage(buffer);
        prev = buffer->slots + prev_diff;
        // last added slot in shm new page is linked to the prev->next
        // in old queue
        buffer->slots[buffer->known_slots - 1].next = prev->next;

        bp_log(FNC_LOG_DEBUG,
               "BPSlots page added in SHM memory (%u slots - %s)\n",
               buffer->known_slots, buffer->filename);
        break;
    case buff_local:
        prev_diff = prev - buffer->slots;
        if (!(added = realloc(buffer->slots, (buffer->control->nslots +
                              BPBUFF_MEM_PAGE) * sizeof(BPSlot))))
            return NULL;
        buffer->slots = added;

        prev = buffer->slots + prev_diff;
        // init new slots
        for (i = buffer->control->nslots;
             i < buffer->control->nslots + BPBUFF_MEM_PAGE - 1; i++) {
            added[i].refs = 0;
            added[i].slot_seq = 0;
            added[i].next = i + 1;
        }
        // last slot
        added[i].refs = 0;
        added[i].slot_seq = 0;
        added[i].next = prev->next;

        added = &added[buffer->control->nslots];
        buffer->control->nslots += BPBUFF_MEM_PAGE;
        buffer->known_slots = buffer->control->nslots;

        bp_log(FNC_LOG_DEBUG,
               "BPSlots page added in local bufferpool (%u slots)\n",
               buffer->known_slots);
        break;
    default:
        return NULL;
        break;
    }

    prev->next = BPtoSlotPtr(buffer, added);

    return added;
}

void bp_free(BPBuffer * buffer)
{
    switch (buffer->type) {
    case buff_shm:
        bp_shm_unmap(buffer);
        bp_log(FNC_LOG_DEBUG, "Buffer in SHM unmapped \n");
        break;
    case buff_local:
        sem_destroy(&buffer->control->syn);
        free(buffer->control);
        free(buffer->slots);
        free(buffer);
        bp_log(FNC_LOG_DEBUG, "Buffer is freed \n");
        break;
    default:
        break;
    }

}

/**
 * @return the next slot in the buffer and do NOT move current read position.
 * NULL if buffer is empty
 */

BPSlot *bp_getreader(BPConsumer * cons)
{
    BPSlot *last_read;
    BPSlot *next;

    bp_lock(cons->buffer);

    if (bp_shm_refresh(cons->buffer)) {
        bp_unlock(cons->buffer);
        return NULL;
    }

    last_read = BPtoSlot(cons->buffer, cons->last_read_pos);
    next = &cons->buffer->slots[cons->read_pos];

    if (!next->refs || (next->slot_seq < cons->last_seq)) {
        // added some slots?
        if (last_read && cons->buffer->slots[last_read->next].refs &&
            (cons->buffer->slots[last_read->next].slot_seq >
             cons->last_seq))
            next = &cons->buffer->slots[last_read->next];
        else {
            bp_unlock(cons->buffer);
            return NULL;
        }
    } else if (last_read
           && (cons->buffer->slots[last_read->next].slot_seq <
               next->slot_seq))
        next = &cons->buffer->slots[last_read->next];

    /* Some stats for the network scheduler */
    next->last_timestamp = cons->buffer->last_slot_timestamp;
    next->pkt_num = cons->buffer->last_slot_seq - next->slot_seq + 1;

    bp_unlock(cons->buffer);
    return next;
}

/**
 * @brief get a new empty slot from bufferpool.
 *
 * It does NOT mark the slot as busy, this action will be performed by
 * bp_write that MUST be called at the end of operations with the obtained
 * slot.
 *
 * @return empty slot.
 */

BPSlot *bp_getslot(BPBuffer * buffer)
{
    BPSlot *slot;

    bp_lock(buffer);

    if (bp_shm_refresh(buffer))
        return NULL;
    
    slot = &buffer->slots[buffer->control->write_pos];

    if (buffer->slots[slot->next].refs > 0) {
        if (!(slot = bp_addpage(buffer, slot))) {
            bp_unlock(buffer);
            return NULL;
        }
    } else {
        slot = &buffer->slots[slot->next];
    }

    bp_unlock(buffer);

    return slot;
}

#define RETURN_ERR    do { \
                bp_unlock(cons->buffer); \
                return 1; \
            } while (0)

/**
 * move current read position
 * @return 1 on error, 0 otherwise
 */

int bp_gotreader(BPConsumer * cons)
{
    BPSlot *last_read;
    BPSlot *next;

    bp_lock(cons->buffer);

    if (bp_shm_refresh(cons->buffer))
        RETURN_ERR;

    last_read = BPtoSlot(cons->buffer, cons->last_read_pos);
    next = &cons->buffer->slots[cons->read_pos];

    if (!next->refs || (next->slot_seq < cons->last_seq)) {
        // added some slots?
        if (last_read && cons->buffer->slots[last_read->next].refs
            && (cons->buffer->slots[last_read->next].slot_seq >
            cons->last_seq))
            next = &cons->buffer->slots[last_read->next];
        else
            RETURN_ERR;
    } else if (last_read
           && (cons->buffer->slots[last_read->next].slot_seq <
               next->slot_seq))
        next = &cons->buffer->slots[last_read->next];

    next->refs--;

    cons->last_seq = next->slot_seq;

//      if ( msync(next, sizeof(BPSlot), MS_ASYNC) )
//      if ( msync(cons->buffer->slots, cons->buffer->known_slots * sizeof(BPSlot), MS_ASYNC) )
//              printf("*** slot msync error\n");

    cons->last_read_pos = BPtoSlotPtr(cons->buffer, next);
    cons->read_pos = next->next;

    // cons->read_pos->refs--;
    bp_dump(cons, NULL);
    bp_unlock(cons->buffer);

    // return cons->read_pos;
    return 0;        // next;
}

#undef RETURN_ERR

#define BPNEWBUFF_RET_ERR \
    do { \
            free(slots); \
            free(buffer); \
            free(control); \
            return NULL; \
        } while (0)

BPBuffer *bp_new(uint32_t buffer_size)
{
    BPSlot *slots = NULL;
    BPBuffer *buffer = NULL;
    BPControl *control = NULL;
    uint32_t index;

    if (!buffer_size)
        return NULL;

    if (!(slots = (BPSlot *) calloc(buffer_size, sizeof(BPSlot))))
        BPNEWBUFF_RET_ERR;
    // *** slots initialization
    for (index = 0; index < buffer_size - 1; index++)
        (slots[index]).next = index + 1;
    /*end of the list back to the head */
    (slots[index]).next = 0;

    // control struct allocation
    if (!(control = malloc(sizeof(BPControl))))
        BPNEWBUFF_RET_ERR;

    control->write_pos = buffer_size - 1;

    control->refs = 0;
    control->nslots = buffer_size;

    if ( (sem_init(&control->syn, 0, 1)) )
        BPNEWBUFF_RET_ERR;

    if (!(buffer = (BPBuffer *) malloc(sizeof(BPBuffer))))
        BPNEWBUFF_RET_ERR;

    buffer->type = buff_local;
    *buffer->filename = '\0';
    buffer->known_slots = buffer_size;

    // link all allocated structs
    buffer->slots = slots;
    buffer->control = control;

    return buffer;
}

#undef BPNEWBUFF_RET_ERR

char *bp_ipc_name(const char *firstname, const char *lastname)
{
    char *dir, *dst, *slash;

    if ((dst = malloc(PATH_MAX)) == NULL)
        return (NULL);

    /* 4can override default directory with environment variable */
    if ((dir = getenv("PX_IPC_NAME")) == NULL) {
#ifdef    POSIX_IPC_PREFIX
        dir = POSIX_IPC_PREFIX;    /* from "config.h" */
#else
        dir = "";    // "/tmp/";   /* default */
#endif
    }
    /* 4dir must end in a slash */
    slash = (strlen(dir) && (dir[strlen(dir) - 1] == '/')) ? "" : "/";
    snprintf(dst, PATH_MAX, "%s%s%s.%s", dir, slash, firstname, lastname);

    return (dst);        /* caller can free() this pointer */
}

#define RETURN(x)    do { \
                bp_unlock(cons->buffer); \
                return x; \
            } while (0)

/**
 * @brief Check if buffer is empty for a given consumer. 
 *
 * Checks in this function are taken from <tt>bp_read()</tt> and not
 * optimized
 *
 * @param cons Consumer to be checked.
 *
 * @retval 1 Buffer is empty.
 * @retval 0 Buffer not empty.
 * @retval -1 Error.
 *
 * @see bp_read
 */

int bp_isempty(BPConsumer * cons)
{
    BPSlot *last_read;
    BPSlot *next;

    bp_lock(cons->buffer);

    if(bp_shm_refresh(cons->buffer))
        RETURN(-1);

    last_read = BPtoSlot(cons->buffer, cons->last_read_pos);
    next = &cons->buffer->slots[cons->read_pos];

    if (!next->refs || (next->slot_seq < cons->last_seq)) {
        // added some slots?
        if (last_read && cons->buffer->slots[last_read->next].refs
            && (cons->buffer->slots[last_read->next].slot_seq >
            cons->last_seq))
            RETURN(0);
        else
            RETURN(1);
    } else if (last_read
           && (cons->buffer->slots[last_read->next].slot_seq <
               next->slot_seq))
        RETURN(0);

    bp_unlock(cons->buffer);

    return 0;
}

#undef RETURN


/**
 * @brief return the next packet timestamp
 *
 * @return -1 on error, the packet timestamp otherwise
 */

double bp_nextts(BPConsumer * cons)
{
    BPSlot *last_read;
    BPSlot *next;
    int ret = -1;

    bp_lock(cons->buffer);

    if(bp_shm_refresh(cons->buffer))
        goto err_unlock;

    last_read = BPtoSlot(cons->buffer, cons->last_read_pos);
    next = &cons->buffer->slots[cons->read_pos];

    if (!next->refs || (next->slot_seq < cons->last_seq)) {
        // added some slots?
        if (last_read && cons->buffer->slots[last_read->next].refs
            && (cons->buffer->slots[last_read->next].slot_seq >
            cons->last_seq))
            next = &cons->buffer->slots[last_read->next];
        else {
            goto err_unlock;
        }
    } else if (last_read
           && (cons->buffer->slots[last_read->next].slot_seq <
               next->slot_seq))
        next = &cons->buffer->slots[last_read->next];
    
    ret = next->timestamp;

    err_unlock:
    bp_unlock(cons->buffer);

    return ret;
}

/*! Read the next slot in buffer.
 *
 * Read the next slot in the buffer fill the parameters given as
 * input and move current read position.
 *
 * @param cons is the consumer.
 * @param timestamp return position for timestamp value of slot read.
 * @param marker return position for marker value of slot read.
 * @param data position of where to copy data filed of the read slot.
 * @param data_size input value for size of <tt>data</tt> buffer and return
 * value for size of effective size of read data.
 *
 * @return -1 on error, 1 if slot data size is bigger than the buffer
 * provided and then it was not possible to copy all the data, 0 otherwise.
 * 
 **/

int bp_read(BPConsumer *cons, uint32_t *timestamp, uint8_t *marker,
           uint8_t * data, uint32_t* data_size)
{
    BPSlot *last_read;
    BPSlot *next;
    uint32_t cpy_size;
    int ret = -1; // wrong by default

    bp_lock(cons->buffer);

    if(bp_shm_refresh(cons->buffer))
        goto err_unlock;

    //we use BPtoSlot 'cause last_read could be NULL
    last_read = BPtoSlot(cons->buffer, cons->last_read_pos);
    next = &cons->buffer->slots[cons->read_pos];

    if (!next->refs || (next->slot_seq < cons->last_seq)) {
        // added some slots?
        if (last_read && cons->buffer->slots[last_read->next].refs
            && (cons->buffer->slots[last_read->next].slot_seq >
            cons->last_seq))
            next = &cons->buffer->slots[last_read->next];
        else {
            goto err_unlock;
        }
    } else if (last_read
           && (cons->buffer->slots[last_read->next].slot_seq <
               next->slot_seq))
        next = &cons->buffer->slots[last_read->next];

    cpy_size = omsbuff_min(*data_size, next->data_size);

    next->refs--;

    cons->last_seq = next->slot_seq;
    // cons->read_pos = next;

    cons->read_pos = next->next;

    // cons->read_pos->refs--;
    // fill input parameters
    *timestamp = next->timestamp;
    *marker = next->marker;
    memcpy(data, next->data, cpy_size);
    *data_size = cpy_size;

    cons->last_read_pos = BPtoSlotPtr(cons->buffer, next);

    ret = (cpy_size == next->data_size) ? 0 : 1;

    err_unlock:
    bp_unlock(cons->buffer);

    return ret;    // next;          
}


/**
 * Add and return a new consumer reference to the buffer,
 * @return NULL if an error occurs
 */

BPConsumer *bp_ref(BPBuffer * buffer)
{
    BPConsumer *cons;

    if (!buffer)
        return NULL;

    if ((cons = (BPConsumer *) malloc(sizeof(BPConsumer))) == NULL)
        return NULL;

    cons->last_read_pos = -1;    // BPtoSlotPtr(buffer, NULL);
    cons->buffer = buffer;
    cons->frames = 0;

    bp_lock(buffer);
    cons->read_pos = buffer->slots[buffer->control->write_pos].next;
    cons->last_seq = buffer->slots[buffer->control->write_pos].slot_seq;

    buffer->control->refs++;

    bp_unlock(buffer);

    bp_log(FNC_LOG_DEBUG, "Buffer ref (%d)\n", buffer->control->refs);

    return cons;
}

void bp_unref(BPConsumer * cons)
{
    if (cons) {
        bp_lock(cons->buffer);
        if (cons->buffer->control->refs > 0) {
            --(cons->buffer->control->refs);
            bp_unlock(cons->buffer);
            //Now consumer has to read all unread slots
            while (!bp_gotreader(cons));
        } else
            bp_unlock(cons->buffer);

        bp_log(FNC_LOG_DEBUG, "Buffer ref (%d)\n",
            cons->buffer->control->refs);
        free(cons);
    }
}

/**
 * @brief Write a new slot in the buffer using the input parameters.
 *
 * The function writes a new slot according with input parameters. If the
 * <tt>data</tt> belongs to a slot that was previously requested with
 * <tt>bp_getslot</tt> then there is not copy of data, just the
 * parameter setting like a commit of changes done.
 * @param seq_delta is the difference between sequence number of pkt(slot) and
 * the expected number. If it's 0, it does mean that the sequence number is
 * grown by 1.
 * @return ERR_NOERROR or ERR_ALLOC
 */

int bp_write(BPBuffer * buffer, int16_t seq_delta, double timestamp,
                  uint32_t rtp_time, uint8_t marker, uint8_t * data, 
                  uint32_t data_size)
{
    BPSlot *slot; // = &buffer->slots[buffer->control->write_pos];
    uint64_t curr_seq; // = slot->slot_seq;
    uint32_t data_to_copy = 
        (data_size < BPSLOT_DATASIZE) ? data_size : BPSLOT_DATASIZE;

    bp_lock(buffer);

    if (bp_shm_refresh(buffer))
        return ERR_ALLOC;

    slot = &buffer->slots[buffer->control->write_pos];
    curr_seq = slot->slot_seq;

    if (buffer->slots[slot->next].data == data) {
        slot = &buffer->slots[slot->next];
    } else {
        if (buffer->slots[slot->next].refs > 0) {
            if ((slot = bp_addpage(buffer, slot)) == NULL) {
                bp_unlock(buffer);
                return ERR_ALLOC;
            }
        } else {
            slot = &buffer->slots[slot->next];
        }

        memcpy(slot->data, data, data_to_copy);
        slot->data_size = data_to_copy;
        slot->timestamp = timestamp;
        slot->rtp_time = rtp_time;
        slot->marker = marker;
        slot->seq_delta = seq_delta + 1;
    }

    buffer->last_slot_timestamp = slot->timestamp;
    slot->refs = buffer->control->refs;
    buffer->last_slot_seq = slot->slot_seq = curr_seq + 1;

    buffer->control->write_pos = BPtoSlotPtr(buffer, slot);

    bp_dump(NULL, buffer);

    bp_unlock(buffer);

    return ERR_NOERROR;
}
