/* *
 * This file is part of bufferpool
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * bufferpool is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * bufferpool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with bufferpool; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

/**
 * @file bp_shm.c
 * Shared memory bufferpool
 */

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "bp_internal.h"
#include "bufferpool.h"


/*!@brief This function adds shared memory page of slots.
 * The function remaps the shared memory according to the new size, but
 * it does NOT link new page to previous existing queue:
 * This is done in bp_addpage.
 * WARNING: the function assumes that the caller (bp_write or bp_getslot) locked the buffer mutex
 * @return the first BPSlot of new added page of slots.
 * */
BPSlot *bp_shm_addpage(BPBuffer * buffer)
{
    BPSlot *added;
    BPSlot *slots;
    unsigned int i;
    char *shm_file_name;
    int fd;
    struct stat fdstat;

    // *** slots mapping in shared memory ***
    if (!(shm_file_name = bp_ipc_name(buffer->filename, BPBUFF_SHM_SLOTSNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR, 0);
    free(shm_file_name);
    if ((fd < 0)) {
        bp_log(FNC_LOG_ERR,
               "Could not open POSIX shared memory (BPSlots):"
               "is Felix running?\n");
        return NULL;
    }
    if ((fstat(fd, &fdstat) < 0)) {
        bp_log(FNC_LOG_ERR, "Could not stat %s\n",
            BPBUFF_SHM_SLOTSNAME);
        close(fd);
        return NULL;
    }

    if (((size_t) fdstat.st_size !=
         buffer->control->nslots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
               "Strange size for shared memory!"
               "(not the number of slots reported in control struct)\n");
        close(fd);
        return NULL;
    }
    if (ftruncate(fd,
            (buffer->control->nslots + BPBUFF_SHM_PAGE) * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
               "Could not set correct size for shared memory object"
               "(BPControl)\n");
        close(fd);
        return NULL;
    }
    if (munmap(buffer->slots, buffer->known_slots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR, "Could not unmap previous slots!!!\n");
        close(fd);
        return NULL;
    }
    slots = mmap(NULL,
                 (buffer->control->nslots + BPBUFF_SHM_PAGE) * sizeof(BPSlot),
                 PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    close(fd);
    if (slots == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        return NULL;
    }
    // inizialization of BPSlots added
    for (i = buffer->control->nslots;
         i < buffer->control->nslots + BPBUFF_SHM_PAGE - 1; i++) {
        slots[i].refs = 0;
        slots[i].slot_seq = 0;
        slots[i].next = i + 1;
    }
    slots[i].refs = 0;
    slots[i].slot_seq = 0;
    // last added slot in shm new page has next slot to NULL:
    // bp_addpage will link it correctly.
    slots[i].next = -1;

    added = &slots[buffer->control->nslots];
    buffer->slots = slots;
    buffer->control->nslots += BPBUFF_SHM_PAGE;
    buffer->known_slots = buffer->control->nslots;

    return added;
}

#define FILE_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)

/**
 * @brief This function creates che shared memory object on the file system.
 *  In order to let different, and independent, processes to access it.
 * WARNING: this function is not used by feng directly but it is
 * implemented here for completeness.
 * Feng is just the "consumer" of a shared memory object,
 * while the producer is some other object like felix.
 * @todo unlink on error
 */

BPBuffer *bp_shm_create(char *shm_name, uint32_t buffer_size)
{
    BPBuffer *buffer;
    BPSlot *slots;
    BPControl *control;
    int fd, shm_open_errno;
    uint32_t index;
    char *shm_file_name;

    if (!buffer_size)
        return NULL;

    // *** control struct shared memory object ***
    if (!(shm_file_name = bp_ipc_name(shm_name, BPBUFF_SHM_CTRLNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR | O_CREAT | O_EXCL, FILE_MODE);
    shm_open_errno = errno;
    if ((fd < 0)) {
        switch (shm_open_errno) {
        case EEXIST:
            bp_log(FNC_LOG_ERR,
                "SHM object \"%s\" already exists!"
                " Perhaps some other apps are using it\n",
                shm_file_name);
            bp_log(FNC_LOG_ERR,
                "TIP: If you are sure none is using"
                " it try deleting it manually.\n");
            break;
        case EINVAL:
            bp_log(FNC_LOG_ERR,
                "Invalid name (%s) was given for shared memory object\n",
                shm_file_name);
            break;
        case EACCES:
            bp_log(FNC_LOG_ERR,
                "Permission denied for shared memory object %s\n",
                shm_file_name);
            break;
        case ENOENT:
            bp_log(FNC_LOG_ERR, "Could not create %s\n",
                shm_file_name);
            break;
        default:
            break;
        }
        bp_log(FNC_LOG_ERR,
            "Could not open/create POSIX shared memory %s (BPControl)\n",
            shm_file_name);
        free(shm_file_name);
        return NULL;
    }
    free(shm_file_name);
    if (ftruncate(fd, sizeof(BPControl))) {
        bp_log(FNC_LOG_ERR,
            "Could not set correct size for shared memory object"
            " (BPControl)\n");
        close(fd);
        return NULL;
    }
    control = mmap(NULL, sizeof(BPControl),
                   PROT_READ | PROT_WRITE, MAP_SHARED,
                   fd, 0);
    close(fd);
    if (control == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        return NULL;
    }

    // inizialization of BPControl
    //control->syn = sem_open(semaphore, O_CREAT, 0644, 1);
    if ( sem_init(&control->syn, 1, 1) ) {
        bp_log(FNC_LOG_FATAL, "Unable to open semaphore for shared memory\n");
        return NULL;
    }

    sem_wait(&control->syn);
    control->refs = 0;
    control->nslots = buffer_size;

    // *** slots mapping in shared memory ***
    if (!(shm_file_name = bp_ipc_name(shm_name, BPBUFF_SHM_SLOTSNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR | O_CREAT | O_EXCL, FILE_MODE);
    shm_open_errno = errno;
    free(shm_file_name);
    if ((fd < 0)) {
        switch (shm_open_errno) {
        case EEXIST:
            bp_log(FNC_LOG_ERR,
                "SHM object \"%s\" already exists!"
                " Perhaps some other apps are using it\n",
                shm_file_name);
            bp_log(FNC_LOG_ERR,
                "TIP: If you are sure none is using it try deleting"
                " it manually.\n");
            break;
        case EINVAL:
            bp_log(FNC_LOG_ERR,
                "Invalid name (%s) was given for shared memory object\n",
                shm_file_name);
            break;
        default:
            break;
        }
        bp_log(FNC_LOG_ERR,
            "Could not open/create POSIX shared memory (BPSlots)\n");
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    if (ftruncate(fd, buffer_size * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
            "Could not set correct size for shared memory object (BPControl)\n");
        close(fd);
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    slots = mmap(NULL, buffer_size * sizeof(BPSlot),
                 PROT_READ | PROT_WRITE, MAP_SHARED,
                 fd, 0);
    close(fd);
    if (slots == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    // inizialization of BPSlots
    for (index = 0; index < buffer_size - 1; index++) {
        slots[index].refs = 0;
        slots[index].slot_seq = 0;
        slots[index].next = index + 1;    // &(slots[index+1]) - slots;
    }
    // slots - slots; /*end of the list back to the head*/
    slots[index].next = 0;
    slots[index].slot_seq = 0;
    // last slot not yet initialized if for cycle.
    slots[index].refs = 0;

    control->write_pos = buffer_size - 1;

    if (!(buffer = (BPBuffer *) malloc(sizeof(BPBuffer)))) {
        munmap(slots, control->nslots * sizeof(BPSlot));
        munmap(control, sizeof(BPControl));
        return NULL;
    }

    buffer->type = buff_shm;
    buffer->slots = slots;
    buffer->known_slots = control->nslots;
    strncpy(buffer->filename, shm_name, sizeof(buffer->filename) - 1);
    buffer->slots = slots;
    buffer->control = control;

    sem_post(&control->syn);

    return buffer;
}

/*! @brief This function destroys the shared memory object.
 * Firstly it unlinks the shm objects referenced in buffer and
 * then unlink them thus removing them from system.
 * It doesn't stop on the first error, but completes all operations
 * and returns the code of first error.
 * @return 0 on succes or the number of the first error occurred.
 * */

int bp_shm_destroy(BPBuffer * buffer)
{
    int unmap_err, shm_unlink_err = 0;
    char *shm_file_name;

    unmap_err = bp_shm_unmap(buffer);

    if (!(shm_file_name = bp_ipc_name(buffer->filename, BPBUFF_SHM_CTRLNAME)))
        return 1;
    if (shm_unlink(shm_file_name)) {
        switch (errno) {
        case ENOENT:
            bp_log(FNC_LOG_ERR, "SHM Object %s doesn't exists\n",
                   shm_file_name);
            break;
        case EACCES:
            bp_log(FNC_LOG_ERR, "Permission denied on SHM Object %s\n",
                   shm_file_name);
            break;
        default:
            bp_log(FNC_LOG_ERR, "Could not unlink SHM Object %s\n",
                   shm_file_name);
            break;
        }
        shm_unlink_err = errno;
    }
    free(shm_file_name);

    if (!(shm_file_name = bp_ipc_name(buffer->filename, BPBUFF_SHM_SLOTSNAME)))
        return 1;
    if (shm_unlink(shm_file_name)) {
        switch (errno) {
        case ENOENT:
            bp_log(FNC_LOG_ERR, "SHM Object %s doesn't exists\n",
                   shm_file_name);
            break;
        case EACCES:
            bp_log(FNC_LOG_ERR, "Permission denied on SHM Object %s\n",
                   shm_file_name);
            break;
        default:
            bp_log(FNC_LOG_ERR, "Could not unlink SHM Object %s\n",
                   shm_file_name);
            break;
        }
        shm_unlink_err = shm_unlink_err ? shm_unlink_err : errno;
    }
    free(shm_file_name);

    return unmap_err ? unmap_err : shm_unlink_err;
}

BPBuffer *bp_shm_map(char *shm_name)
{
    BPBuffer *buffer;
    BPSlot *slots;
    BPControl *control;
//      uint32_tindex;
    char *shm_file_name;
    int fd;
    struct stat fdstat;

    // *** control struct mapping in shared memory ***
    if (!(shm_file_name = bp_ipc_name(shm_name, BPBUFF_SHM_CTRLNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR, 0);
    free(shm_file_name);
    if ((fd < 0)) {
        bp_log(FNC_LOG_ERR,
            "Could not open POSIX shared memory (BPControl):"
            " is Felix running?\n");
        return NULL;
    }
    if ((fstat(fd, &fdstat) < 0)) {
        bp_log(FNC_LOG_ERR, "Could not stat %s\n",
            BPBUFF_SHM_CTRLNAME);
        close(fd);
        return NULL;
    }

    if (((size_t) fdstat.st_size != sizeof(BPControl))) {
        bp_log(FNC_LOG_ERR,
            "Strange size for BPControl shared memory!"
            " (not an integer number of slots)\n");
        close(fd);
        return NULL;
    }
    control = mmap(NULL, fdstat.st_size,
                   PROT_READ | PROT_WRITE, MAP_SHARED,
                   fd, 0);
    close(fd);
    if (control == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        return NULL;
    }
    // *** XXX inizialization of control struct made by creator!
    // control->refs = 0;
    // control->nslots = buffer_size-1;

    // if ( pthread_mutexattr_init(&mutex_attr) )
    //      return NULL;
    // if ( pthread_mutex_init(&control->syn, &mutex_attr) )
    //      return NULL;
    // *** wait for completion af all operations, especially for initializzation
    sem_wait(&control->syn);
    sem_post(&control->syn);

    // *** slots mapping in shared memory ***
    if (!(shm_file_name = bp_ipc_name(shm_name, BPBUFF_SHM_SLOTSNAME)))
        return NULL;

    fd = shm_open(shm_file_name, O_RDWR, 0);
    free(shm_file_name);
    if ((fd < 0)) {
        bp_log(FNC_LOG_ERR,
            "Could not open POSIX shared memory (BPSlots):"
            " is Felix running?\n");
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    if ((fstat(fd, &fdstat) < 0)) {
        bp_log(FNC_LOG_ERR, "Could not stat %s\n", BPBUFF_SHM_SLOTSNAME);
        close(fd);
        munmap(control, sizeof(BPControl));
        return NULL;
    }

    if (((size_t) fdstat.st_size != control->nslots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
               "Strange size for shared memory!"
               " (not an integer number of slots)\n");
        close(fd);
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    slots = mmap(NULL, fdstat.st_size,
                 PROT_READ | PROT_WRITE, MAP_SHARED,
                 fd, 0);
    close(fd);
    if (slots == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        munmap(control, sizeof(BPControl));
        return NULL;
    }
    // initialized by creator!
    // for(index=0; index<buffer->nslots-1; index++)
    //      (slots[index]).next = BPSLOT_SETADDR(buffer, &(slots[index+1]));
    // (slots[index]).next=BPSLOT_SETADDR(buffer, slots); /*end of the list back to the head*/

    if (!(buffer = (BPBuffer *) malloc(sizeof(BPBuffer)))) {
        munmap(slots, control->nslots * sizeof(BPSlot));
        munmap(control, sizeof(BPControl));
        return NULL;
    }

    buffer->type = buff_shm;
    buffer->slots = slots;
    buffer->known_slots = control->nslots;
    strncpy(buffer->filename, shm_name, sizeof(buffer->filename) - 1);
    // buffer->fd = -1;
    // buffer->fd = NULL;
    buffer->slots = slots;
    buffer->control = control;

    return buffer;
}

/*! @brief This function remaps shared memory slots in case that the slots
 * queue size is changed.
 * The function only remap the shared memory according to the new size, but
 * it does not do anything on new shm slots page: if it must be initialized
 * you must do it by yourself.
 * WARNING: the function assumes that the caller locked the buffer mutex.
 * @return 0 on succes, 1 otherwis.
 * */

int bp_shm_remap(BPBuffer * buffer)
{
    BPSlot *slots;
    char *shm_file_name;
    int fd;
    struct stat fdstat;

    // *** slots mapping in shared memory ***
    if (!(shm_file_name = bp_ipc_name(buffer->filename, BPBUFF_SHM_SLOTSNAME)))
        return 1;

    fd = shm_open(shm_file_name, O_RDWR, 0);
    free(shm_file_name);
    if ((fd < 0)) {
        bp_log(FNC_LOG_ERR,
               "Could not open POSIX shared memory (BPSlots):"
               " is Felix running?\n");
        return 1;
    }
    if ((fstat(fd, &fdstat) < 0)) {
        bp_log(FNC_LOG_ERR, "Could not stat %s\n",
            BPBUFF_SHM_SLOTSNAME);
        close(fd);
        return 1;
    }

    if (((size_t) fdstat.st_size !=
         buffer->control->nslots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR,
               "Strange size for shared memory!"
               " (not the number of slots reported in control struct)\n");
        close(fd);
        return 1;
    }

    if (munmap(buffer->slots, buffer->known_slots * sizeof(BPSlot))) {
        bp_log(FNC_LOG_ERR, "Could not unmap previous slots!!!\n");
        close(fd);
        return 1;
    }
    slots = mmap(NULL, fdstat.st_size,
                 PROT_READ | PROT_WRITE, MAP_SHARED,
                 fd, 0);
    close(fd);
    if (slots == MAP_FAILED) {
        bp_log(FNC_LOG_FATAL, "SHM: error in mmap\n");
        return 1;
    }

    buffer->slots = slots;
    buffer->known_slots = buffer->control->nslots;

    bp_log(FNC_LOG_DEBUG, "SHM memory remapped (known slots %d)\n",
        buffer->known_slots);

    return 0;
}

int bp_shm_unmap(BPBuffer * buffer)
{
    int slots_err;
    int control_err;

    if (buffer->type != buff_shm) {
        bp_log(FNC_LOG_ERR,
               "Bufferpool desn't seems to be a Shared Memory object");
        return 1;
    }

    if ((slots_err =
         munmap(buffer->slots, buffer->control->nslots * sizeof(BPSlot))))
        bp_log(FNC_LOG_ERR, "Error unmapping BPSlots SHM object\n");
    if ((control_err = munmap(buffer->control, sizeof(BPControl))))
        bp_log(FNC_LOG_ERR, "Error unmapping BPControl SHM object\n");

    return slots_err ? slots_err : control_err;
}
